"""How much rain is trapped in Codelandia?

No buildings mean no rain is captured::

    >>> rain([])
    0

All-same height buildings capture no rain::

    >>> rain([10])
    0

    >>> rain([10, 10])
    0

    >>> rain([10, 10, 10, 10])
    0

If there's nothing between taller buildings, no rain is captured::

    >>> rain([2, 3, 10])
    0

    >>> rain([10, 3, 2])
    0

If two tallest buildings are same height and on ends, it's easy::

    >>> rain([10, 5, 10])
    5

    >>> rain([10, 2, 3, 4, 10])
    21

    >>> rain([10, 4, 3, 2, 10])
    21

    >>> rain([10, 2, 4, 3, 10])
    21

If two tallest buildings are ends, but not the same height,
it will fall off the shorter of thh two::

    >>> rain([10, 2, 3, 4, 9])
    18

Rain falls off the left and right edges::

    >>> rain([2, 3, 10, 5, 5, 10, 3, 2])
    10

Trickier::

    >>> rain([2, 3, 5, 4, 3, 10, 7, 10, 5, 4, 3, 6, 2, 5, 2])
    15

Should also work with floats::

    >>> r = rain([4.5, 2.2, 2.2, 4])
    >>> round(r, 2)
    3.6
"""


def rain(buildings):
    """How much rain is trapped in Codelandia?"""

    # START SOLUTION

    rain = 0

    # Special case: no buildings means nothing trapped

    if not buildings:
        return rain

    for i, height in enumerate(buildings):

        # Find highest building to our left (including us)
        left_best = max(buildings[:i + 1])

        # Find highest building to our right (including us)
        right_best = max(buildings[i:])

        # The shorter of these is "leak level"; rain is our height to that

        leak_level = min(left_best, right_best)
        rain += leak_level - height

    return rain


# ADVANCED: Linear O(n) solution

def highest_so_far(buildings):
    """Returns list of highest-building-seen-so-far for each building."

    For example::

        >>> buildings = [1, 3, 2, 5, 2, 1]

    This is "list of 'highest building to my left'"::

        >>> highest_so_far(buildings)
        [1, 3, 3, 5, 5, 5]

    You can use this to get rightward answer by reversing input & result
    (this is "list of 'highest building to my right'")::

        >>> list(reversed(highest_so_far(reversed(buildings))))
        [5, 5, 5, 5, 2, 1]
    """

    so_far = 0
    out = []

    for height in buildings:
        so_far = max(height, so_far)
        out.append(so_far)

    return out


def rain_linear(buildings):
    """How much rain is trapped in Codelandia?"""

    rain = 0

    # Special case: no buildings means nothing trapped

    if not buildings:
        return rain

    # The amount of rain caught by a building is defined by
    # its highest-building-ever-on-the-left and
    # its highest-building-ever-on-the-right.

    # Make lists of highest-left and highest-right

    highest_on_left = highest_so_far(buildings)
    highest_on_right = reversed(highest_so_far(reversed(buildings)))

    for height, best_l, best_r in zip(buildings, highest_on_left, highest_on_right):
        leak_level = min(best_l, best_r)
        rain += leak_level - height

    return rain


def rain_linear_alternate(buildings):
    """How much rain is trapped in Codelandia?"""

    if not buildings:
        return 0

    top = max(buildings)  # height of tallest building

    runoff = 0

    # amount of water that runs off to the left
    highest = 0
    for height in buildings:
        if height == top:  # no more water will drain to the left
            break
        highest = max(highest, height)
        runoff += top - highest

    # amount of water that runs off to the right
    highest = 0
    for height in reversed(buildings):
        if height == top:
            break
        highest = max(highest, height)
        runoff += top - highest

    # trapped water = area of bounding box - area of buildings - total runoff

    return top * len(buildings) - sum(buildings) - runoff

# END SOLUTION

if __name__ == '__main__':
    import doctest
    if doctest.testmod().failed == 0:
        print("\n*** ALL TESTS PASSED. YAY!\n")
